# Copyright (c) 2017  University of Houston

import os
import sys
import math
import time
import json
import logging
import argparse
import shutil
import numpy as np
import pandas as pd
from gplotter import GPlotter
from pcap2csv import makecsvs

logi = logging.getLogger('traffic_metrics')
logi.setLevel(logging.DEBUG)

class Traffic_Metrics():
    
    def __init__(self, logh=None):
        self.out_dir = None
        self.plots_dir = None
        self.plotter = None
        self.plot_graphs = False
        self.logger = logh if logh else logi
        #metric-funtion-map
        self.mf_map = {
            'general':{
                'duration':     ['secs', self.getDuration],
                'total_packets':['packets', self.getTotalPackets],
                'packet_rate':  ['pps', self.getPacketRate],
                'inter_packet_time':['secs', self.getIPT],
                'dropped_packets':['packets', self.getDropped]
            },
            'l2':{
                'l2_total_payload_bytes': ['bytes', self.getL2TotalPayload],
                'l2_payload_throughput':['bps', self.getL2PayloadThroughput],
                'l2_payload_size':['bytes', self.getL2PayloadSize],
                'l2_total_vlan_packets':['packets',self.getTotalVLANPackets],
                'l2_#packets_per_vlan':['packets', self.getPacketsPerVLAN],
                'l2_number_of_vlans':['vlans', self.getNumVLANs]
            },
            'l3':{
                'l3_total_payload_bytes':['bytes', self.getL3TotalPayload],
                'l3_payload_throughput':['bps', self.getL3PayloadThroughput],
                'l3_payload_size':['bytes', self.getL3PayloadSize],
                'l3_fragmented_packets':['packets',self.getTotalFragPackets]
            },
            'l4':{
                'l4_total_payload_bytes':['bytes', self.getL4TotalPayload],
                'l4_payload_throughput':['bps', self.getL4PayloadThroughput],
                'l4_payload_size':['bytes', self.getL4PayloadSize],
                'l4_total_retransmitted_pkts':['packets',self.getTotalRetransPkts],
                'l4_total_outoforder_pkts':['packets', self.getTotalOutOfOrder],
                'l4_retransmitted_rate':['pps', self.getRetransPktRate],
                'l4_outoforder_rate':['pps', self.getOutOfOrderRate],
                'l4_window_size':['bytes', self.getWindowSize],
                'l4_unacked_inflight_data':['bytes', self.getUnackedData],
                'l4_total_acks':['packets', self.getTotalAcks],
                'l4_total_push':['packets', self.getTotalPush]
            }
        }
        self.pcap2_metrics = ['dropped_packets']
        self.conn_pkts_reqd_metrics = ['l4_unacked_inflight_data']

        #graphable metrics
        self.graph_metrics = ['packet_rate', 'inter_packet_time', 'dropped_packets',
                'l2_payload_throughput', 'l2_payload_size','l3_payload_throughput',
                'l3_payload_size', 'l4_payload_throughput', 'l4_payload_size',
                'l4_retransmitted_rate', 'l4_outoforder_rate', 'l4_window_size',
                'l4_unacked_inflight_data']
        #stats metrics (calculate mean median mode)
        self.stat_metrics = ['packet_rate', 'inter_packet_time', 'dropped_packets',
                'l2_payload_throughput', 'l2_payload_size','l3_payload_throughput',
                'l3_payload_size', 'l4_payload_throughput', 'l4_payload_size',
                'l4_retransmitted_rate', 'l4_outoforder_rate', 'l4_window_size',
                'l4_unacked_inflight_data']

    def setOutDir(self, out_dir):
        out_dir = os.path.abspath(out_dir)
        assert(os.path.isdir(out_dir)), "the output directory supplied  %s is \
                                        not a valid directory".format(out_dir)
        self.out_dir = out_dir
    
    def makeEmptyPlotsDir(self):
        self.plots_dir = self.out_dir + '/plots'
        if os.path.exists(self.plots_dir):
            shutil.rmtree(self.plots_dir)
        os.makedirs(self.plots_dir)

    def setPlotter(self):
        self.makeEmptyPlotsDir()
        self.plotter = GPlotter(self.plots_dir)
    
    def t_start(self, pkt_list):
        return pkt_list.iloc[0].time
        
    def t_end(self, pkt_list):
        return pkt_list.iloc[-1].time
    
    def getStartStopTimes(self, pkt_lists):
        start = self.t_start(pkt_lists[0])
        stop = self.t_end(pkt_lists[0])
        
        for pkt_list in pkt_lists:
            start = min(start, self.t_start(pkt_list))
            stop = max(start, self.t_end(pkt_list))
        return (start, stop)
    
    def filterPackets(self, pkt_list, filter_dict):
        #TODO implement other filters 
        if filter_dict == {}:
            return pkt_list
        cond = True
        if 'time' in filter_dict:
            cond = cond & (pkt_list['time'] >= filter_dict['time'][0])
            cond = cond & (pkt_list['time'] <= filter_dict['time'][1])
        if 'src_mac' in filter_dict:
            cond = cond & (pkt_list['src_mac'] == filter_dict['src_mac'])
        if 'dst_mac' in filter_dict:
            cond = cond & (pkt_list['dst_mac'] == filter_dict['dst_mac'])
        if 'src_ip' in filter_dict:
            cond = cond & (pkt_list['src_ip'] == filter_dict['src_ip'])
        if 'dst_ip' in filter_dict:
            cond = cond & (pkt_list['dst_ip'] == filter_dict['dst_ip'])
        if 'src_port' in filter_dict:
            cond = cond & (pkt_list['src_port']==int(filter_dict['src_port']))
        if 'dst_port' in filter_dict:
            cond = cond & (pkt_list['dst_port']==int(filter_dict['dst_port']))
        #print( cond)
        return pkt_list[cond] 
    
    def getXpsList(self, df_t_val):
        '''
        return a list of packets per seconds
        '''

        df_t_val = df_t_val[np.isfinite(df_t_val['val'])]   
        #print( 'ppppppppppp\n', df_t_val.val)
        if len(df_t_val) == 0:
            return [[], []]
        dp = 1 #decimal places for interval
        dec, step = 10**dp, 1./(10**dp)
        start, stop = self.getStartStopTimes([df_t_val])
        xps_times = (np.trunc(np.arange(start, stop + step,
                                        step)*dec)/dec).tolist()
        groups = df_t_val.groupby(pd.cut(df_t_val.time, xps_times))
        if groups.count().empty:
            return [[], []]
        #TODO:make pcap2csv return valid l2_d_size for arp_packets/other layer2
        last_int = df_t_val.iloc[-1].time - groups.first().iloc[-1].time
        starts = groups.first().time
        widths = groups.first().time.diff().shift(-1)
        widths.iloc[-1] = last_int
        xps_list = groups.val.sum()

        widths = widths.replace(0, step)
        widths = widths.fillna(step)
        xps_list = xps_list*1/widths
        xps_list = xps_list.fillna(0)
        return starts.tolist(), xps_list.tolist()

    def getStatistics(self, my_list, unit):
        if len(my_list) == 0:
            my_mean, my_std, my_min, my_max = [0,0,0,0]
        else:
            ar = np.array(my_list)
            my_mean = np.mean(ar)
            my_std = np.std(ar)
            my_min = min(my_list) if len(my_list) > 0 else None
            my_max = max(my_list) if len(my_list) > 0 else None
        return {'mean':my_mean, 
                'std':my_std,
                'max':my_max,
                'min':my_min,
                'unit':unit
        }
    
    def plotGraphs(self, x, y, prefix, metric_name, unit):
        
        self.plotter.plotline(x, y, prefix + ' - ' + metric_name, unit)
        self.plotter.plotdistr(y, prefix + ' - ' + metric_name +\
                               ' distribution', unit)
        self.plotter.plotCDF(y, prefix + ' - ' + metric_name +' cdf')

    def processList(self, x, y, prefix, metric_name, unit):

        if metric_name in self.graph_metrics and self.plot_graphs:
            self.plotGraphs( x, y, prefix, metric_name, unit)
        if metric_name in self.stat_metrics:
            return self.getStatistics(y, unit)

    ####general
    def getDuration(self, name, unit,  prefix, l_pkt_list):
        return {'value':self.t_start(l_pkt_list) - \
                        self.t_end(l_pkt_list),
                'unit':unit}

    def getTotalPackets(self, name, unit,  prefix, l_pkt_list):
        return {'value':len(l_pkt_list),
                'unit':unit}

    def getPacketRate(self, name, unit,  prefix, l_pkt_list):
        '''
        return a list of packets per seconds
        '''
        l_pkt_list.loc[:,'val'] = 1
        x,y = self.getXpsList(l_pkt_list)
        return self.processList( x, y, prefix, name, unit)

    def getIPT(self, name, unit,  prefix, l_pkt_list):
        '''
        return a list ipt per seconds bfill first
        '''
        if len(l_pkt_list)<=1:
            return [[], []]
        x,y = l_pkt_list.time.tolist(), \
                     l_pkt_list.time.diff().fillna(method='bfill').tolist()
        return self.processList( x, y, prefix, name, unit)
    
    def getDropped(self, name, unit,  prefix, l_pkt_list, l_pkt_list2):
        return {'value':len(l_pkt_list) - len(l_pkt_list2),
        'unit':unit}

    ####l2
    def getL2TotalPayload(self, name, unit,  prefix, l_pkt_list):
        return {'value':l_pkt_list.l2_d_size.sum(),
                'unit':unit}

    def getL2PayloadThroughput(self, name, unit,  prefix, l_pkt_list):
        '''
        return a list of l2 bits per seconds of payload
        '''
        l_pkt_list.loc[:, 'val'] = l_pkt_list.l2_d_size * 8
        l_pkt_list['val'].fillna(0)
        x,y =  self.getXpsList(l_pkt_list)
        return self.processList( x, y, prefix, name, unit)

    def getL2PayloadSize(self, name, unit,  prefix, l_pkt_list):
        '''
        return a list of l2 pkt payload sizes. 
        replace Nan seen in the pcap with 0
        '''
        if len(l_pkt_list)==0:
            return [[], []]
        x,y = l_pkt_list.time.tolist(), l_pkt_list.l2_d_size.fillna(0).tolist()
        return self.processList( x, y, prefix, name, unit)

    def getTotalVLANPackets(self, name, unit,  prefix, l_pkt_list):
        return {'value': len(l_pkt_list[l_pkt_list.vlan_id.notnull()]),
               'unit':unit}
    
    def getPacketsPerVLAN(self, name, unit,  prefix, l_pkt_list):
        ret = {'unit':unit}
        for l_key, v_pkt_list in l_pkt_list.groupby(['vlan_id']):
            ret[l_key] = len(v_pkt_list)
        return ret
    
    def getNumVLANs(self, name, unit,  prefix, l_pkt_list):
        return {'value': len (l_pkt_list.vlan_id[l_pkt_list.vlan_id.notnull() \
                              ].unique()),
               'unit':unit}

    ####l3
    def getL3TotalPayload(self, name, unit,  prefix, l_pkt_list):
        return {'value':l_pkt_list.l3_d_size.sum(),
                'unit':unit}

    def getL3PayloadThroughput(self, name, unit,  prefix, l_pkt_list):
        '''
        return a list of l3 bits per seconds of payload
        '''
        l_pkt_list.loc[:,'val'] = l_pkt_list.l3_d_size * 8
        l_pkt_list['val'].fillna(0)
        x,y =  self.getXpsList(l_pkt_list)
        return self.processList( x, y, prefix, name, unit)

    def getL3PayloadSize(self, name, unit,  prefix, l_pkt_list):
        '''
        return a list of l3 pkt payload sizes. 
        replace Nan seen in the pcap with 0
        '''
        if len(l_pkt_list)==0:
            return [[], []]
        x,y = l_pkt_list.time.tolist(), l_pkt_list.l3_d_size.fillna(0).tolist()
        return self.processList( x, y, prefix, name, unit)

    def getTotalFragPackets(self, name, unit,  prefix, l_pkt_list):
        return {'value': len (l_pkt_list.defrag_pkt_idx[\
                               l_pkt_list.defrag_pkt_idx.notnull()]),
               'unit':unit}

    ####l4
    def getL4TotalPayload(self, name, unit,  prefix, l_pkt_list):
        return {'value':l_pkt_list.d_size.sum(),
                'unit':unit}

    def getL4PayloadThroughput(self, name, unit,  prefix, l_pkt_list):
        '''
        return a list of l4 bits per seconds of payload
        '''
        l_pkt_list.loc[:,'val'] = l_pkt_list['d_size'] * 8
        l_pkt_list['val'].fillna(0)
        x,y =  self.getXpsList(l_pkt_list)
        return self.processList( x, y, prefix, name, unit)

    def getL4PayloadSize(self, name, unit,  prefix, l_pkt_list):
        '''
        return a list of l2 pkt payload sizes. 
        replace Nan seen in the pcap with 0
        '''
        if len(l_pkt_list)==0:
            return [[], []]
        x,y = l_pkt_list.time.tolist(), l_pkt_list.d_size.fillna(0).tolist()
        return self.processList( x, y, prefix, name, unit)

    def getTotalRetransPkts(self, name, unit,  prefix, l_pkt_list):
        ret_list = l_pkt_list.loc[l_pkt_list['is_retransmission'] == 1]
        return {'value':len(ret_list),
                'unit':unit}

    def getTotalOutOfOrder(self, name, unit,  prefix, l_pkt_list):
        ret_list = l_pkt_list.loc[l_pkt_list['is_out_of_order'] == 1]
        return {'value':len(ret_list),
                'unit':unit} 

    def getRetransPktRate(self, name, unit,  prefix, l_pkt_list):
        '''get retransmitted pkts per second in TCP pkts'''
        l_pkt_list.loc[:,'val'] = l_pkt_list['is_retransmission']
        x,y = self.getXpsList(l_pkt_list)
        return self.processList( x, y, prefix, name, unit)

    def getOutOfOrderRate(self, name, unit,  prefix, l_pkt_list):
        '''get retransmitted pkts per second in TCP pkts'''
        l_pkt_list.loc[:,'val'] = l_pkt_list['is_out_of_order']
        x,y = self.getXpsList(l_pkt_list)
        return self.processList( x, y, prefix, name, unit)

    def getWindowSize(self, name, unit,  prefix, l_pkt_list):
        if len(l_pkt_list) == 0:
            return [[], []]
        start, stop = self.getStartStopTimes([l_pkt_list])
        winsizetimelist = l_pkt_list[['time', 'win_size']].dropna()
        times = winsizetimelist['time'].tolist()
        winsizes = winsizetimelist['win_size'].tolist()
        x,y = times, winsizes
        return self.processList( x, y, prefix, name, unit)

    #TODO:
    def getUnackedData(self, name, unit,  prefix, l_pkt_list, conn_pkt_list):
        #amount of inflight unacked data per time
        #ignores sacks
        src_side = prefix.split('_')[-2] + '_' + prefix.split('_')[-1]
        times, unackeddata = [], []
        last_real_ack_num = -1
        for i, pkt in conn_pkt_list.iterrows():
            #print( pkt)
            s1 = pkt.src_ip + '_' + str(pkt.src_port)
            #print( s1, src_side, )
            if s1 == src_side:
                #print( True)
                times.append(pkt.time)
                inflight = pkt.seq_num +  pkt.d_size - last_real_ack_num \
                           if last_real_ack_num != -1 else 0
                unackeddata.append(inflight)
            else:
                #print( False)
                if pkt.ack == 1:
                    last_real_ack_num = pkt.ack_num
        x,y = times, unackeddata
        return self.processList( x, y, prefix, name, unit)
    
    def getTotalAcks(self, name, unit,  prefix, l_pkt_list):
        ret_list = l_pkt_list.loc[(l_pkt_list['ack'] == 1) & 
                                  (l_pkt_list['d_size'] == 0)]
        return {'value':len(ret_list),
                'unit':unit}

    def getTotalPush(self, name, unit,  prefix, l_pkt_list):
        ret_list = l_pkt_list.loc[(l_pkt_list['psh'] == 1)]
        return {'value':len(ret_list),
                'unit':unit}

    def analyzeLayer(self, l_groups, l_groups2, layer, prefix, 
                     conn_pkt_list=None):
        ret = {}
        for l_key, l_pkt_list in l_groups:
            print( "......packets of ", l_key)
            ret[l_key] = {}
            mprefix = prefix + '_' + l_key
            metrics = dict(self.mf_map['general'], **self.mf_map[layer])
            for name, (unit, funct) in metrics.iteritems():
                #print( name, self.pcap2_metrics, type(l_groups2))
                if name in self.conn_pkts_reqd_metrics:
                    ret[l_key][name] = funct(name, unit, mprefix, l_pkt_list, 
                                             conn_pkt_list)
                elif name in self.pcap2_metrics:
                    if l_groups2:
                        #print( 'True')
                        l_pkt_list2 = l_groups2.get_group(l_key)
                        ret[l_key][name] = funct (name, unit, mprefix, l_pkt_list,
                                                l_pkt_list2)
                else:
                    ret[l_key][name] = funct(name, unit, mprefix, l_pkt_list)
        return ret



        self.pcap2_metrics = ['dropped_packets']
        self.conn_pkts_reqd_metrics = ['l4_unacked_data']

    def analyzeL2(self, pkt_list, pkt_list2):
        print( '...Working on layer 2 analysis')
        ret = {}
        prefix = 'Layer2'
        #global - all L2 packets analysis
        l_key = 'all_l2_pkts'
        ret[l_key] = {}
        mprefix = prefix + '_' + l_key
        metrics = dict(self.mf_map['general'], **self.mf_map['l2'])
        pcap2isDF = isinstance(pkt_list2, pd.DataFrame)
        for name, (unit, funct) in metrics.iteritems():
            if name in self.pcap2_metrics:
                if pcap2isDF:
                    ret[l_key][name] = funct (name, unit, mprefix, pkt_list,
                                            pkt_list2)
            else:
                ret[l_key][name] = funct (name, unit, mprefix, pkt_list)
        #per mac pair analysis
        pkt_list['l2_key'] = pkt_list['src_mac'] + '-' + pkt_list['dst_mac']
        l2_groups = pkt_list.groupby(['l2_key'])
        #if pkt_list2 != None: 
        if pcap2isDF:
            pkt_list2['l2_key'] =pkt_list2['src_mac'] +'-'+pkt_list2['dst_mac']
        #l2_groups2 = pkt_list2.groupby(['l2_key']) if pkt_list2 else None
        l2_groups2 = pkt_list2.groupby(['l2_key']) if pcap2isDF else None
        retb = self.analyzeLayer(l2_groups, l2_groups2, 'l2', prefix)
        ret.update(retb)
        return ret

    def analyzeL3(self, pkt_list, pkt_list2):
        print( '...Working on layer 3 analysis')
        pkt_list['l3_key'] = pkt_list['src_ip'] + '-' + pkt_list['dst_ip']
        l3_groups = pkt_list.groupby(['l3_key'])
        pcap2isDF = isinstance(pkt_list2, pd.DataFrame)
        if pcap2isDF:
            pkt_list2['l3_key'] = pkt_list2['src_ip'] + '-' +pkt_list2['dst_ip']
        l3_groups2 = pkt_list2.groupby(['l3_key']) if pcap2isDF else None
        prefix = 'Layer3'
        return self.analyzeLayer(l3_groups, l3_groups2, 'l3', prefix)
                
    def analyzeL4(self, pkt_list, pkt_list2):
        print( '...Working on layer 4 analysis')
        ret = {}
        conns_grp1 = pkt_list.groupby(['sess_idx'])
        pcap2isDF = isinstance(pkt_list2, pd.DataFrame)
        conns_grp2 = pkt_list2.groupby(['sess_idx']) if pcap2isDF else None
        #print( conns_grp1.count())
        #print( conns_grp1.count())
        for conn_id, cpkt_list in conns_grp1:
            print( '......Working on connection', conn_id)
            if len(cpkt_list) == 0:
                print( '.........Connection', conn_id, 'has no packets-skipping')
                continue
            cpkt_list2 = conns_grp2.get_group(conn_id) if pcap2isDF else None

            cpkt_list.loc[:,'l4_src_key'] = cpkt_list['src_ip'] + '_' + \
                                      cpkt_list['src_port'].map(str)
            cpkt_list.loc[:,'l4_dst_key'] = cpkt_list['dst_ip'] + '_' + \
                            cpkt_list['dst_port'].map(str)
            ret[conn_id] = {}
            ret[conn_id]['type'] = cpkt_list.iloc[0].l4_type
            ret[conn_id]['conn_key'] = cpkt_list.iloc[0].l4_src_key + '__' +\
                                       cpkt_list.iloc[0].l4_dst_key
            ret[conn_id]['duration'] =cpkt_list.time.max()-cpkt_list.time.min()

            #do calculations per side of each connection
            if pcap2isDF:
                cpkt_list2.loc[:,'l4_src_key'] = cpkt_list2['src_ip'] + '_' + \
                                           cpkt_list2['src_port'].map(str)
            c_side_grps = cpkt_list.groupby(['l4_src_key'])
            c_side_grps2 = cpkt_list2.groupby(['l4_src_key']) if \
                                pcap2isDF else None
            prefix = ret[conn_id]['type'] + '_conn_' + str(conn_id) + '_side'
            per_side_dict = self.analyzeLayer(c_side_grps, c_side_grps2, 'l4',
                                              prefix, cpkt_list)
            #already calculated duration at global connection level
            for side_dict in per_side_dict.values():
                del side_dict['duration']
            ret[conn_id].update(per_side_dict)
        return ret

    def getPktsDataframe(self, pkts_dataset_file, filter_dict):
        init_pkt_list = pd.read_csv(pkts_dataset_file)
        st = self.t_start(init_pkt_list)
        init_pkt_list['time_stamp'] = init_pkt_list['time']
        init_pkt_list['time'] = init_pkt_list['time'] - st
        return self.filterPackets(init_pkt_list, filter_dict)

    def analyze(self, pcap1, out_dir, pcap2=None, filter_str=None,
                analysis_type='all', plot_graphs=False, keep_dataset=False):
        report_dict = {'filter': filter_str}
        filter_dict = json.loads(filter_str) if filter_str else {}
        self.setOutDir(out_dir)
        self.setPlotter()
        self.plot_graphs = plot_graphs
        makecsvs(open(pcap1), self.out_dir+'/pcap1_dataset','packets')
        pkt_list = self.getPktsDataframe(self.out_dir+'/pcap1_dataset/dataset/'\
                                         'all_pkts.csv', filter_dict)
        if pcap2:
            makecsvs(open(pcap2), self.out_dir+'/pcap2_dataset/',
                              'packets')
            pkt_list2 = self.getPktsDataframe(self.out_dir+'/pcap1_dataset/'\
                                              'dataset/all_pkts.csv', filter_dict)
        else: 
            pkt_list2 = None
        pg = plot_graphs
        if analysis_type in ['layer2', 'all']:
            report_dict['layer2'] = self.analyzeL2(pkt_list, pkt_list2) 
        if analysis_type in ['layer3', 'all']: 
            report_dict['layer3'] = self.analyzeL3(pkt_list, pkt_list2)
        if analysis_type in ['layer4', 'all']: 
            report_dict['layer4'] = self.analyzeL4(pkt_list, pkt_list2)
        with open(out_dir + '/report.json', 'w') as fo:
            json.dump(report_dict, fo, indent=2)
        if not keep_dataset:
            shutil.rmtree(self.out_dir+'/pcap1_dataset')
            if pcap2:
                shutil.rmtree(self.out_dir+'/pcap2_dataset')
        print('\nplots are stored in', out_dir + '/plots')
        print( 'report is at', out_dir + '/report.json\n')
